type format =
  [ `Altergo
  | `Dimacs
  | `Smtlib2 ]

let format = ref `Extension

let set_format = function
  | "alt-ergo" -> format := `Altergo
  | "dimacs"   -> format := `Dimacs
  | "smtlib2"  -> format := `Smtlib2
  | _ -> assert false

let dimacs f =
    match Dimacs.check_file f with
    | Dimacs.Sat -> Format.printf "Sat@."
    | Dimacs.Unsat -> Format.printf "Unsat@."

let altergo f =
  let parsed = Popop_of_altergo.read_file f in
  match Popop_of_altergo.check_goal parsed with
  | Popop_of_altergo.Sat | Popop_of_altergo.Idontknow ->
    Format.printf "I don't know@."
  | Popop_of_altergo.Valid ->
    Format.printf "Valid@."

let smtlib2 f =
  let cmds = Popop_of_smtlib2.read_file f in
  let res = Popop_of_smtlib2.check_file cmds in
  Format.printf "%a@." Popop_of_smtlib2.print_status res

let get_extension file =
  let s = try Filename.chop_extension file
    with Invalid_argument _ -> invalid_arg "unspecified extension" in
  let n = String.length s + 1 in
  String.sub file n (String.length file - n)

let args = [Debug.Args.desc_debug_list;
            Debug.Args.desc_debug;
            Debug.Args.desc_debug_all;
            ("--lang", Arg.Symbol(["alt-ergo";"dimacs";"smtlib2"], set_format),
               "Set the input format (default use the extension)");
            ("--seed", Arg.Int (fun i -> Shuffle.set_shuffle (Some [|i|])),
             "For debugging set the seed used")
           ]

let get_format filename =
  match !format with
  | #format as x -> x
  | `Extension ->
    let ext = get_extension filename in
    match ext with
    | "mlw" | "why" -> `Altergo
    | "cnf" -> `Dimacs
    | "smt2" -> `Smtlib2
    | _ -> Printf.printf "Unknwon extension %s\n" ext; exit 1

let do_file f =
  match get_format f with
  | `Altergo -> altergo f
  | `Dimacs -> dimacs f
  | `Smtlib2 -> smtlib2 f

let files =
  let files = Queue.create () in
  Arg.parse args (fun f -> Queue.add f files)
    "popop [file.why|file.cnf|file.smt2]";
  Debug.Args.set_flags_selected ();
  if Debug.Args.option_list () then exit 0;
  files

let () =
  if not (Solver.check_initialization ()) ||
     not (Conflict.check_initialization ()) then
    exit 1

let () =
  try
    Queue.iter do_file files
  with e when not (Debug.test_flag Debug.stack_trace) ->
    Format.eprintf "%a@." Exn_printer.exn_printer e;
    exit 1
