(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)
open Popop
open Stdlib
open Types
open Explanation


exception DontKnowThisCon
exception DontKnowThisDec of Dom.K.t * Cl.t

exception DomNeverSet
exception ClNotEqual

let debug = Debug.register_info_flag
  ~desc:"for@ the@ main@ part@ of@ conflicts."
  "Conflict.core"

type 'a rescon = 'a Explanation.rescon =
| GRequested: 'a -> 'a rescon
| GOther: 'b con * 'b ->  'a rescon

let return : type a b. a con -> b con -> b -> a rescon = fun con con' v ->
  match Explanation.Con.Eq.eq_type con con' with
  | Some Types.Eq -> GRequested v
  | None -> GOther(con',v)

type explearnt =
| ExpLearnt: Tags.t -> explearnt

let explearnt : explearnt exp = Exp.create_key "Explanation.learnt_exp"
let dom_bool : bool dom = Dom.create_key "bool"

exception NotMarkedExp

(*
module type SemH = sig
  type k
  type v
  module S : Types.Sem with type t = k
  val h : v S.H.t
end

module SemH = Sem.MkVector(struct
    type ('a,'b) t = (module SemH with type k = 'a and type v = 'b)
  end)
*)

(** *)
let hash_chogen = ref (fun _ -> assert false)
let equal_chogen = ref (fun _ _ -> assert false)

module ChoH = Stdlib.XHashtbl.Make(struct
    type t = chogen
    let hash chogen  = !hash_chogen chogen
    let equal chogen1 chogen2 = !equal_chogen chogen1 chogen2
  end)

(** Needed *)
(*
module AgeCache = struct
  module KeyCache = Types.Make_key(struct end)
  module Cache = KeyCache.MkVector(struct type ('a,'b) t = 'a end)
  type t = unit Cache.t Age.M.t

  let no_cache : unit KeyCache.k = KeyCache.create_key "no cache"

  let new_cache () = Cache.create (KeyCache.hint_size ())

  let empty = Age.M.empty
  let copy m = Age.M.map (fun _ -> new_cache ()) m
  let add_age m age = Age.M.add_new Impossible age (new_cache ()) m

  let get_cache m age =
    match Age.M.find_smaller_opt age m with
    | None -> assert false (** The age must be added in advance then used *)
    | Some (_,c) -> c

  let is_cached c key = not (Cache.is_uninitialized c key)
  let get_cached c key =
    if Cache.is_uninitialized c key
    then None else Some (Cache.get c key)
  let set_cached c key v =
    if KeyCache.equal key no_cache then () else
    Cache.set c key v

  let nb_hit = Debug.register_stats_int ~name:"Rlist.Cache.hit" ~init:0
  let nb_miss = Debug.register_stats_int ~name:"Rlist.Cache.miss" ~init:0

  let memoise f agecache key age x =
    let agecache = get_cache agecache age in
    match get_cached agecache key with
    | None ->
      Debug.incr nb_miss;
      let r = f age x in
      (** f can run for the same key and an age in the same interval *)
      if Cache.is_uninitialized agecache key then
        set_cached agecache key r
      else
        Debug.dprintf0 debug "[memoise] @[memoise: cycle?@]@\n";
      r
    | Some r ->
      Debug.incr nb_hit;
      r

  type 'a cache = 'a KeyCache.k
  let create_cache = KeyCache.create_key

end
*)
type edge = {
  eto : Cl.t; (** equals to *)
  epexp : pexp (** explication for this merge *);
  einv  : bool; (** which way was pexp *)
}

let print_edge fmt e =
    Format.fprintf fmt "M(%a,@,%a,@,%b)"
      Cl.print e.eto !print_pexp e.epexp e.einv

type rlist = edge list

let print_rlist fmt l = Pp.print_list Pp.semi print_edge fmt l

(** Con_iter *)
type con_iter = {
  csolver : Solver.t; (** state before the last dec *)
  csolver_unsat : Solver.t;
  (** state at unsat (* TODO remove just for first_sem *)*)
  mutable cdeps : Explanation.Deps.t;
  ctrail : Explanation.t;
  cinvclhist: Explanation.invclhist;
  cdomcho: chogen Cl.H.t Dom.Vector.t;
  cdomage: Age.t Cl.H.t Dom.Vector.t;
  cseen  : Explanation.concache Simple_vector.t; (** to clear at the end *)
}

let create_con_iter solver solver_unsat trail =
  let con_iter = {
    csolver  = solver;
    csolver_unsat  = solver_unsat;
    cdeps    = Deps.empty;
    ctrail   = trail;
    cinvclhist = Explanation.invclhist trail;
    cseen    = Simple_vector.create 0;
    cdomcho  = Dom.Vector.create (Dom.hint_size ());
    cdomage  = Dom.Vector.create (Dom.hint_size ());
  } in
  (*
    Debug.dprintf4 debug "@[invclhist:%a@\ndomhist:%a@]@\n"
    Explanation.print_invclhist con_iter.cinvclhist
    Explanation.print_domhist trail.domhist;
  *)
  (** initialization of cdomage *)
  Dom.iter {Dom.iter = (fun dom ->
      Dom.Vector.set con_iter.cdomage dom (Cl.H.create 10))};
  (** initialization of ccdomcho *)
  Dom.iter {Dom.iter = (fun dom ->
      Dom.Vector.set con_iter.cdomcho dom (Cl.H.create 10))};
  con_iter

let confact : unit con = Con.create_key "Core.fact"
let conclause: bool Cl.M.t Explanation.con =
  Explanation.Con.create_key "Bool.clause"



let return_rescon (type a) (type b) (con : a con) (x:a)
    (conrequested : b con) : b rescon =
  match Con.Eq.eq_type con conrequested with
  | None    -> (GOther (con,x) : b rescon)
  | Some Types.Eq -> GRequested x


let mk_confact con = return_rescon confact () con

let return_nothing con = return con confact ()

let _print_dom_can_be_at_current fmt v =
  Dom.Vector.print Pp.newline Pp.space
    {Dom.Vector.printk = Dom.print}
    (Pp.print_iter2 Cl.H.iter Pp.semi Pp.comma Cl.print DBool.print) fmt v


type testconflict =
  | False (* false in this context *)
  | True (* true in this context *)
  | ToDecide (* can be decided *)


type 'a fold_decisions =
  {fold_decisions : 'k 'd. ('a -> ('k,'d) cho -> 'k -> 'd -> 'a) }

class type finalized = object
  method print: Format.formatter -> unit
  method conflict_add: Solver.Delayed.t -> bool Cl.M.t
  method test: Solver.Delayed.t -> testconflict
  method decide: 'a. 'a fold_decisions -> Solver.Delayed.t -> 'a -> 'a
    (** to call only when test is [Todecide] *)
end

let print_finalized fmt fin = fin#print fmt

(** Con *)
module type Con' = sig

  type t

  val print: t Pp.printer

  val same_sem:
    con_iter -> age -> 'a sem -> 'a ->
    pexp -> Cl.t -> Cl.t -> t rescon

  val propacl : con_iter -> Age.t -> Cl.t -> Cl.t -> t rescon

  val finalize: con_iter -> t Bag.t -> finalized option

end

module VCon = Con.MkVector
  (struct type ('a,'unedeed) t =
            (module Con' with type t = 'a)
   end)

let defined_con : unit VCon.t = VCon.create 8

let get_con k =
  assert (if VCon.is_uninitialized defined_con k
    then raise UnregisteredKey else true);
  VCon.get defined_con k

let print_con (type a) (con : a con) fmt (k:a) =
  let module C = (val get_con con) in
  C.print fmt k

let print_rescon (type a) (con : a con) fmt (r:a rescon) =
  match r with
  | GRequested k    -> print_con con  fmt k
  | GOther (cono,k) -> Format.fprintf fmt "%a of %a"
                         (print_con cono) k
                         Con.print cono


(** Exp con Registration *)
module type Exp' = sig

  type t

  val print: t Pp.printer

  (* val iterexp: exp_iter -> age -> t -> unit *)

  val analyse  : con_iter -> age ->
    'a con -> t -> 'a rescon

  val expdom : con_iter -> age ->
    'b dom -> Cl.t ->
    'a con -> t -> 'a rescon

end

module VExp = Exp.MkVector
  (struct type ('a,'unedeed) t =
            (module Exp' with type t = 'a)
   end)

let defined_exp : unit VExp.t = VExp.create 8

let get_exp k =
  assert (if VExp.is_uninitialized defined_exp k
    then raise UnregisteredKey else true);
  VExp.get defined_exp k

let print_pexp fmt (Pexp(age,k,exp,_tag,_)) =
  let f (type b) k exp =
    let s = get_exp k in
    let module S = (val s : Exp' with type t = b) in
    Format.fprintf fmt "exp %a:%a@ at %a"
      Exp.print k S.print exp Age.print age
  in
  f k exp

let () = Explanation.print_pexp := print_pexp

(** Choice *)
type 'd decdone  =
| DecNo (** no decision to do *)
| DecTodo of 'd (** this decision should be done *)

module type Cho' = sig

  type k
  type d

  module Key  : Stdlib.Datatype with type t = k
  module Data: sig
    type t = d
    val print: t Pp.printer
  end

  val choose_decision:
    Solver.Delayed.t -> Key.t -> Data.t decdone

  val make_decision:
    Solver.Delayed.t -> dec -> Key.t -> Data.t -> unit

  val analyse:
   con_iter -> 'a con -> Key.t -> Data.t -> 'a rescon

end


let print_modif fmt = function
  | Cl (cl1,cl2) -> Format.fprintf fmt "Cl %a  -> %a" Cl.print cl1 Cl.print cl2
  | Dom (cl,dom,_,cl0) -> Format.fprintf fmt "Dom %a for %a(%a)"
      Dom.print dom Cl.print cl0 Cl.print cl;
  | Dec _ -> Format.fprintf fmt "Dec"
  | DomL (cl,dom,v,_,_,cl0) ->
    Format.fprintf fmt "DomL %a for %a(%a) with value %a"
      Dom.print dom Cl.print cl0 Cl.print cl (Pp.print_option_or_default "None"
                                   (Solver.print_dom dom)) v;;

let () = Explanation.print_modif_ref := print_modif

(** Give for explanation that something is the last before the limit.
    So we generalize on that. *)
type explimit =
  | LCl: Cl.t * Cl.t -> explimit
  (** At the limit the first cl is represented by this cl *)
  | LDom: 'a dom * Explanation.mod_dom -> explimit
  (** At the limit this class have this value for this dom *)

let print_explimit fmt = function
  | LCl (cl,rcl) -> Format.fprintf fmt "LCl(%a,%a)" Cl.print cl Cl.print rcl
  | LDom (dom,moddom) ->
    Format.fprintf fmt "LDom(%a,%a)" Dom.print dom
      Explanation.print_mod_dom moddom

let explimit : explimit exp = Exp.create_key "Conflict.explimit"

let is_explimit (Pexp(_,exp,k,_,_)) =
  match Exp.Eq.eq_type explimit exp with
  | Some Types.Eq -> Some (k : explimit)
  | None -> None


module ComputeConflict = struct
  let debug = Debug.register_info_flag
    ~desc:"for@ the@ computation@ part@ of@ conflicts."
    "Conflict.compute"

  type t = con_iter

  let before_first_dec t age = Explanation.before_first_dec t.ctrail age
  let _after_last_dec t age = Age.compare t.ctrail.last_dec age <= 0

  type get_equal_aux =
    | GEANoPath
    | GEAAfterDec of rlist
    | GEABeforeDec of Age.t * Cl.t * rlist

  let get_equal t age cl1 cl2 =
    let rec aux last_dec dest clgraph age prev cl = function
      | [] -> GEANoPath
      | node::l when Cl.equal node.ncl prev ->
        aux last_dec dest clgraph age prev cl l
      | node::l ->
        let r =
          if Cl.equal node.ncl dest
          then GEAAfterDec []
          else
            let edge = Cl.M.find node.ncl clgraph in
            aux last_dec dest clgraph age cl node.ncl edge
        in
        let after = Age.compare last_dec node.nage <= 0 in
        match r with
        | GEANoPath -> aux last_dec dest clgraph age prev cl l
        | GEAAfterDec l when after ->
          GEAAfterDec ({eto=node.ncl;epexp=node.npexp;einv=node.ninv}::l)
        | GEAAfterDec l -> (** limit before -> after *)
          GEABeforeDec (node.nage,node.ncl,l)
        | GEABeforeDec (age,ncl,l) when not after ->
          GEABeforeDec (Age.max node.nage age,ncl,l)
        | GEABeforeDec (age,cln,l) -> (** limit after -> before *)
          let pexp =
            Explanation.mk_pexp_direct
              ~age:age
              explimit (LCl (node.ncl,cln)) in
          GEAAfterDec({eto=node.ncl;epexp=node.npexp;einv=node.ninv}::
                        {eto=cln;epexp=pexp;einv=false}::l)
    in
    if Cl.equal cl2 cl1
    then []
    else
      let clgraph = t.ctrail.clgraph in
      let edge = Cl.M.find cl1 clgraph in
      let r = aux t.ctrail.last_dec cl2 clgraph  age cl1 cl1 edge in
      match r with
      | GEANoPath -> raise ClNotEqual
      | GEABeforeDec (age,cln,l) ->
        let pexp =
          Explanation.mk_pexp_direct
            ~age:age
            explimit (LCl (cl1,cln)) in
        {eto=cln;epexp=pexp;einv=false}::l
      | GEAAfterDec l -> l

  let rec get_repr_at t age cl =
    try
      let (agemerge,cl') = Cl.M.find cl t.ctrail.clhist in
      if Age.compare age agemerge < 0 then cl
      else get_repr_at t age cl'
    with Not_found -> cl

  let rec get_repr_at_hist t age acc cl =
    try
      let (agemerge,cl') = Cl.M.find cl t.ctrail.clhist in
      if Age.compare age agemerge < 0 then cl,acc
      else get_repr_at_hist t age (agemerge::acc) cl'
    with Not_found -> cl,acc

  let get_dom t age cl0 dom =
    let rec go_down ~ismergedone age acc = function
      | Explanation.DomNeverSet -> acc
      | Explanation.DomSet(moddom,n)
        when Explanation.Age.compare
            moddom.Explanation.modage t.ctrail.last_dec < 0 ->
        let pexp =
          Explanation.mk_pexp_direct
            ~age:moddom.Explanation.modage
            explimit (LDom (dom,moddom)) in
        go_down ~ismergedone age ({moddom with modpexp = pexp}::acc) n
      | Explanation.DomSet(moddom,n)
        when Explanation.Age.compare moddom.Explanation.modage age <= 0 ->
        go_down ~ismergedone age (moddom::acc) n
      | Explanation.DomMerge(agemerge,n1,n2) ->
        let applicable = Explanation.Age.compare agemerge age <= 0 in
        assert (not ismergedone || applicable );
        let acc = if applicable
          then go_down ~ismergedone:true age acc n1 else acc in
        go_down ~ismergedone:applicable age acc n2
      | Explanation.DomPreMerge(agemerge,_,n1,n2) ->
        let applicable = Explanation.Age.compare agemerge age <= 0 in
        assert (not ismergedone || applicable );
        let acc = if not ismergedone && applicable
          then go_down ~ismergedone:true age acc n1 else acc in
        go_down ~ismergedone:applicable age acc n2
      | Explanation.DomSet(_,n) ->
        (** The age is not reached, go down further *)
        assert (acc == []);
        assert (not ismergedone);
        go_down ~ismergedone:false age acc n in
    assert (Explanation.Age.compare t.ctrail.last_dec age <= 0);
    let clr = (get_repr_at t age cl0) in
    let domhist_node =
      Cl.M.find_def
        Explanation.DomNeverSet
        clr
        (Dom.Vector.get t.ctrail.domhist dom) in
    (* Format.fprintf (Debug.get_debug_formatter ()) *)
    (*   "@[get_repr_at %a %a = %a@, domhist %a@]@\n" *)
    (*   Age.print age Cl.print cl0 Cl.print clr print_domhist_node
         domhist_node; *)
    let acc = go_down ~ismergedone:false age [] domhist_node in
    if acc == [] then raise DomNeverSet else acc

  let get_dom_path t age cl0 dom =
    let rec go_down age acc hcl hdom =
      match hdom,hcl with
      | Explanation.DomNeverSet, _ -> acc
      | Explanation.DomSet(moddom,n), _
        when Explanation.Age.compare
            moddom.Explanation.modage t.ctrail.last_dec < 0 ->
        let pexp =
          Explanation.mk_pexp_direct
            ~age:moddom.Explanation.modage
            explimit (LDom (dom,moddom)) in
        go_down age ({moddom with modpexp = pexp}::acc) hcl n
      | Explanation.DomSet(moddom,n), _
        when Explanation.Age.compare moddom.Explanation.modage age <= 0 ->
        go_down age (moddom::acc) hcl n
      | Explanation.DomMerge(agemerge,_,n2), _
        when Explanation.Age.compare agemerge age > 0 ->
        (** before age *)
        go_down age acc hcl n2
      | Explanation.DomMerge(agemerge,n1,_), agemerge'::hcl when
          Age.equal agemerge agemerge' ->
        (** merged from not representative *)
        go_down age acc hcl n1
      | Explanation.DomMerge(_,_,n2), _ ->
        go_down age acc hcl n2
      | Explanation.DomPreMerge(_,cl,n1,_), _ ->
        let _,hcl = get_repr_at_hist t age [] cl in
        go_down age acc hcl n1
      | Explanation.DomSet(_,n), _ ->
        assert (acc == []); go_down age acc hcl n in
    assert (Explanation.Age.compare t.ctrail.last_dec age <= 0);
    let clr,clhist = get_repr_at_hist t age [] cl0 in
    let domhist_node =
      Cl.M.find_def
        Explanation.DomNeverSet
        clr
        (Dom.Vector.get t.ctrail.domhist dom) in
    (* Format.fprintf (Debug.get_debug_formatter ()) *)
    (*   "@[get_repr_at %a %a = %a@, domhist %a@]@\n" *)
    (*   Age.print age Cl.print cl0 Cl.print clr print_domhist_node
         domhist_node; *)
    let acc = go_down age [] clhist domhist_node in
    if acc == [] then raise DomNeverSet else acc


  let get_dom_last t age cl0 dom =
    let rec go_down age hcl hdom =
      match hdom,hcl with
      | Explanation.DomNeverSet, _ -> raise DomNeverSet
      | Explanation.DomSet(moddom,_), _
        when Explanation.Age.compare
            moddom.Explanation.modage t.ctrail.last_dec < 0 ->
        let pexp =
          Explanation.mk_pexp_direct
            ~age:moddom.Explanation.modage
            explimit (LDom (dom,moddom)) in
        {moddom with modpexp = pexp}
      | Explanation.DomSet(moddom,_), _
        when Explanation.Age.compare moddom.Explanation.modage age <= 0 ->
        moddom
      | Explanation.DomMerge(agemerge,_,n2), _
        when Explanation.Age.compare agemerge age > 0 ->
        (** before age *)
        go_down age hcl n2
      | Explanation.DomMerge(agemerge,n1,_), agemerge'::hcl when
          Age.equal agemerge agemerge' ->
        (** merged from not representative *)
        go_down age hcl n1
      | Explanation.DomMerge(_,_,n2), _ ->
        go_down age hcl n2
      | Explanation.DomPreMerge(_,cl,n1,_), _ ->
        let _,hcl = get_repr_at_hist t age [] cl in
        go_down age hcl n1
      | Explanation.DomSet(_,n), _ ->
        go_down age hcl n in
    assert (Explanation.Age.compare t.ctrail.last_dec age <= 0);
    let clr,clhist = get_repr_at_hist t age [] cl0 in
    let domhist_node =
      Cl.M.find_def
        Explanation.DomNeverSet
        clr
        (Dom.Vector.get t.ctrail.domhist dom) in
    (* Format.fprintf (Debug.get_debug_formatter ()) *)
    (*   "@[get_repr_at %a %a = %a@, domhist %a@]@\n" *)
    (*   Age.print age Cl.print cl0 Cl.print clr print_domhist_node
         domhist_node; *)
    go_down age clhist domhist_node


  let unknown_con t con a : unit =
    t.cdeps <- Deps.add_unknown_con t.cdeps con a

  let add_deps t deps =
    t.cdeps <- Deps.concat t.cdeps deps


  let get_current_deps t = t.cdeps

  let get_pexp_aux (type a) t pexp (con: a con) : a rescon * Deps.t =
    let Pexp(age,k,exp,tags,concache) = pexp in
    Debug.dprintf4 debug
      "analyse@ pexp %a@ into@ con %a@\n @["
      print_pexp pexp Con.print con;
    assert (Exp.equal k explimit ||
            Exp.equal k expcho ||
            Exp.equal k explearnt ||
            Explanation.Age.compare t.ctrail.last_dec age <= 0);
    let f (type b) (k : b exp) (exp: b) =
      let s = get_exp k in
      let module S = (val s : Exp' with type t = b) in
      let r = S.analyse t age con exp in
      Debug.dprintf2 debug "Done:@,%a@]@\n" (print_rescon con) r;
      t.cdeps <- Deps.add_tags t.cdeps tags;
      r
    in
    if Concache.is_set concache con
    then
      let (r,_ as rdeps) = Concache.get concache con in
      Debug.dprintf2 debug "Cached:@,%a@]@\n" (print_rescon con) r;
      rdeps
    else
      let cdeps_saved = t.cdeps in
      t.cdeps <- Deps.empty;
      let r = f k exp in
      let rdeps = r,t.cdeps in
      Concache.set concache con rdeps;
      Simple_vector.push t.cseen concache;
      t.cdeps <- cdeps_saved;
      rdeps

  let get_pexp_deps :
    type a. con_iter -> pexp -> a con -> a rescon * Deps.t =
    fun t pexp (con: a con) ->
      get_pexp_aux t pexp con
(*      let Pexp(age,k,exp,_,_) = pexp in
      match Exp.Eq.eq_type k expdirect with
      | None -> get_pexp_aux t pexp con
      | Some Types.Eq ->
        match exp with
        | DirectDom(dom,cl) ->
          match get_dom t age cl dom with
          | , _ when Age.compare age agemerge >= 0 ->
            assert false (** Because directdom is possible only when the
                             class is a representative *)
          | _, pexp -> get_pexp_deps t pexp con
*)

  let get_pexp t pexp con =
    let r,deps = get_pexp_deps t pexp con in
    (* Format.fprintf (Debug.get_debug_formatter ()) "("; *)
    add_deps t deps;
    (* Format.fprintf (Debug.get_debug_formatter ()) ")"; *)
    r

  (** add because we will have multiple value *)
  let set_dec_cho t cho k =
    t.cdeps <- Deps.add_chogen t.cdeps (GCho(cho,k))

  let set_dec_cho_cond t dom cl cho k =
    Cl.H.add (Dom.Vector.get t.cdomcho dom) cl (GCho(cho,k))

  let compute_con_iter solver solver_unsat trail pexp =
    let state = create_con_iter solver solver_unsat trail in
    let rescon = get_pexp state pexp conclause in
    begin match rescon with
    | GRequested s -> unknown_con state conclause s
    | GOther (con,c) -> unknown_con state con c
    end;
    (** clear the cache *)
    Simple_vector.iter_initialized Concache.clear state.cseen;
    state

  type iter = con_iter
  let dep_dec (type a) iter (dom: a dom) cl : unit =
    let lchogen =
      Cl.H.find_all (Dom.Vector.get iter.cdomcho dom) cl in
    let lchoage =
      Cl.H.find_all (Dom.Vector.get iter.cdomage dom) cl in
    if lchogen == [] && lchoage == [] then
      raise (DontKnowThisDec (((dom : a Dom.k) :> Dom.K.t),cl));
    iter.cdeps <- List.fold_left Deps.add_chogen iter.cdeps lchogen

  class wrapp (l:finalized list) : finalized = object
    method print fmt = (Pp.print_list Pp.semi (fun fmt c -> c#print fmt)) fmt l
    method test d =
      try
        List.fold_left (fun acc c ->
            match c#test d with
            | False -> acc
            | True -> raise Exit
            | ToDecide -> ToDecide) False l
      with Exit -> True

    method decide:
      'a. 'a fold_decisions -> Solver.Delayed.t -> 'a -> 'a =
        fun f d acc ->
          List.fold_left (fun acc (c:finalized) -> c#decide f d acc) acc l

    method conflict_add d =
      List.fold_left
        (fun acc c -> Cl.M.union (fun _ b1 b2 -> assert (b1 == b2); Some b1)
            (c#conflict_add d) acc) Cl.M.empty l
  end


  let finalize state =
    let fold (type b) acc con l =
      if Con.equal con confact then acc
      else
        let con = get_con con in
        let module C = (val con : Con' with type t = b) in
        match C.finalize state l with
        | None -> acc
        | Some mcl -> mcl::acc
    in
    let deps = Deps.compute_deps state.cdeps in
    let acc =
      Conunknown.fold {Conunknown.fold} [] deps.Deps.unknown in
    match acc with
    | []   -> None,deps.Deps.tags, deps.Deps.decs
    | l    -> Some (new wrapp l), deps.Deps.tags, deps.Deps.decs

end


(*
module IterExp = struct
  let debug = Debug.register_info_flag
      ~desc:"for@ the@ explanation@ part@ of@ conflicts."
      "Conflict.exp"

  exception AgeTooYoung
  exception AgeTooOld
  open Tmp
  let rec get_repr_at age cl = function
    | CMerge(rcl,agemerge,_rpexp,l,_) when Age.compare age agemerge >= 0 ->
      (** get_trl can't fail since it have been completed for this age *)
      get_repr_at age rcl (get_trl l)
    | CRepr | CMerge _ -> cl

  type t = exp_iter

  (** This function must record what we must look for in the trail at
      the current level, or when the modification occured at a lower
      level we must create an explimit. Said differently it record the
      reason of the modification between the conflict and
      the state at the last decision *)

  let check_age t age =
    assert (if Age.compare age t.tage > 0 then raise AgeTooYoung
    else if not (Explanation.at_current_level t.etrail age)
    then raise AgeTooOld
    else true)

  let merge_at_current_level t = function
    | CRepr -> true
    | CMerge(_,agemerge,_,_,_) -> Explanation.at_current_level t.etrail agemerge

  let rec add_needed_cl' t age cl =
    match Cl.M.find_opt cl t.tclhist with
    | Some (agemerge,rcl) when Age.compare age agemerge >= 0 ->
      Debug.dprintf4 debug "[Exp] to look for cl %a at %a@\n"
        Cl.print cl Age.print agemerge;
      if Explanation.at_current_level t.etrail agemerge then begin
        (** We look for the explication in the trail *)
        let l = add_needed_cl t age rcl in
        let l = CMerge(rcl,agemerge,create_po (),init_trl l,AgeCache.empty) in
        Age.H.replace t.lookforcl agemerge l;
        l
      end else begin
        let l = add_needed_cl t age rcl in
        let rcl,agemerge,l =
          match l with
          | CRepr -> rcl,agemerge,l
          | CMerge(_,age,_,_,_)
            when Explanation.at_current_level t.etrail age ->
            rcl,agemerge,l
          (** Not None because it is a previously added explimit*)
          (** replace the head with a one for this cl *)
          | CMerge(rcl,agemerge,pexp,rl,_) ->
            assert ( match Rlist.Tmp.get_opt_po pexp with
                | None -> false
                | Some {em_pexp = Pexp(_,exp,_,_,_)} ->
                  Exp.equal exp explimit);
            rcl,agemerge,Rlist.Tmp.get_trl rl in
        (** rcl is the representative at the limit *)
        let pexp = Explanation.mk_pexp_direct
            ~age:agemerge
            explimit (LCl (cl,rcl)) in
        let pexp = {em_other_cl = rcl; em_other_rl = CRepr;
                    em_pexp = pexp;
                    em_repr_cl = rcl; em_repr_rl = CRepr;
                   } in
        CMerge(rcl,agemerge,init_po pexp,init_trl l,AgeCache.empty)
      end
    | Some (agemerge,rcl) ->
      assert (Age.compare age agemerge < 0);
      assert (Explanation.at_current_level t.etrail agemerge);
      CMerge (rcl, agemerge, create_po (), create_trl (),AgeCache.empty)
    | None -> CRepr

  and complete_needed_cl t age cl l =
    assert (Cl.H.find t.needed.ncl cl == l);
    match l with
    | CRepr -> ()
    | CMerge(rcl,agemerge,rpexp,rl',_) ->
      match get_opt_trl rl' with
      | Some l' -> complete_needed_cl t age rcl l'
      | None when Age.compare age agemerge >= 0 ->
        let l' = add_needed_cl t age rcl in
        Age.H.replace t.lookforcl agemerge l;
        set_trl l l'
      | None -> ()

  and add_needed_cl t age cl =
    try
      let l = Cl.H.find t.needed.ncl cl in
      complete_needed_cl t age cl l;
      l
    with Not_found ->
      let l = add_needed_cl' t age cl in
      assert (not (Cl.H.mem t.needed.ncl cl));
      Cl.H.add t.needed.ncl cl l;
      l

  let need_cl_repr t age cl =
    check_age t age;
    let l = add_needed_cl t age cl in
    Debug.dprintf2 debug "[Exp] @[need_cl_repr: %a@]@\n" Rlist.Tmp.print l


  (** What about none at limit? What is none for a value of a dom *)
  let dom_not_none_at_limit = function
    | None -> raise DomNoneAtLimit
    | Some v -> v

  let save_dom_at_limit t lastlevel rcl dom v =
    let hcl = Dom.Vector.get t.needed.ndom dom in
    let mage = Cl.H.find_def hcl Age.M.empty rcl in
    Debug.dprintf6 debug  "[Exp] @[save dom at limit %a for %a at %a@]@\n"
      Dom.print dom Cl.print rcl Age.print lastlevel;
    let pexp =
        Explanation.mk_pexp_direct
          ~age:lastlevel
          explimit (LDom (dom,rcl,v)) in
    let mage = Age.M.add_new Impossible lastlevel pexp mage in
    Cl.H.add hcl rcl mage

  let dom_at_limit_saved t rcl dom =
    let hcl = Dom.Vector.get t.needed.ndom dom in
    let ages = Cl.H.find_def hcl Age.M.empty rcl in
    Age.M.find_smaller_opt (Explanation.last_dec t.etrail) ages != None

  let need_dom t age cl dom =
    Debug.dprintf6 debug  "[Exp] @[need_dom %a for %a at %a@]@\n"
      Dom.print dom Cl.print cl Age.print age;
    check_age t age;
    let l = add_needed_cl t age cl in
    let rcl = get_repr_at age cl l in
    Debug.dprintf4 debug "[Exp] @[  repr:%a rl: %a@]@\n"
      Cl.print rcl Rlist.Tmp.print l;
    let dom_can_be_at_current_level =
      let hcl = Dom.Vector.get t.dom_can_be_at_current dom in
      try
        let b = Cl.H.find hcl rcl in
        assert (b
                || dom_at_limit_saved t rcl dom
                || raise DomNeverSet );
        b
      with Not_found ->
        let lastlevel = Explanation.last_dom DomNeverSet t.etrail dom rcl in
        Debug.dprintf4 debug "[Exp] @[  lastlevel: %a; rcl: %a@]@\n"
          Age.print lastlevel Cl.print rcl;
        if Explanation.at_current_level t.etrail lastlevel
        then begin Cl.H.add hcl rcl true; true end
        else begin
          (** rcl is the representative at the limit *)
          (** the domain is not at the current level, we do the needing
              book keeping *)
          Cl.H.add hcl rcl false;
          (** since lastlevel is before the current level what is
              currently associated in the state of the solver is the value of
              the domain at the limit for this cl *)
          let v = dom_not_none_at_limit
              (Solver.get_direct_dom t.solver dom rcl) in
          save_dom_at_limit t lastlevel rcl dom v;
          false
        end in
    Debug.dprintf1 debug "[Exp] @[dcbacl:%b@]@\n"
      dom_can_be_at_current_level;
    if dom_can_be_at_current_level then
      t.older.(Age.to_int age) <-
        FDom (rcl,dom)::t.older.(Age.to_int age)


  let need_sem (type a) t age (sem : a sem) v =
    Debug.dprintf6 debug "[Exp] @[need sem %a with value %a at %a@]@\n"
      Sem.print sem (Solver.print_sem sem) v Age.print age;
    check_age t age;
    let module Sem = (val (SemH.get t.needed.nsem sem)) in
    try
      let l = Sem.S.H.find Sem.h v in
      match l with
      | CRepr -> assert false
      | CMerge(rcl,agemerge,rpexp,l',_) when Age.compare age agemerge >= 0 ->
        assert (is_init_trl l');
        Debug.dprintf4 debug "[Exp] @[rcl=%a l=%a@]@\n"
          Cl.print rcl
          Rlist.Tmp.print l;
        complete_needed_cl t age rcl (get_trl l')
      | CMerge _ -> ()
    with Not_found ->
      match Solver.get_sem_first t.solver sem v with
      | None -> raise SemNeverSet
      | Some (rcl,agemerge,pexp) ->
        assert (Age.compare age agemerge >= 0);
        let l =
          if Explanation.at_current_level t.etrail agemerge then begin
            (** We look for the explication in the trail *)
            let l = add_needed_cl t age rcl in
            let l = CMerge(rcl, agemerge, create_po (), init_trl l,
                           AgeCache.empty) in
            Age.H.replace t.lookforsem agemerge l;
            l
          end else begin
            let l = add_needed_cl t age rcl in
            let rcl,agemerge,l =
              match l with
              | CRepr -> rcl,agemerge,l
              | CMerge(_,age,_,_,_)
                when Explanation.at_current_level t.etrail age ->
                rcl,agemerge,l
              (** Not None because it is a previously added explimit*)
              (** replace the head with a one for this sem *)
              | CMerge(rcl,agemerge,pexp,rl,_) ->
                assert ( match Rlist.Tmp.get_opt_po pexp with
                    | None -> false
                    | Some {em_pexp = Pexp(_,exp,_,_,_)} ->
                      Exp.equal exp explimit);
                rcl,agemerge,Rlist.Tmp.get_trl rl in
            (** rcl is the representative at the limit *)
            let pexp = Explanation.mk_pexp_direct
                ~age:agemerge
                explimit (LSem (sem,v,rcl)) in
            let pexp = {em_other_cl = rcl; em_other_rl = CRepr;
                        em_pexp = pexp;
                        em_repr_cl = rcl; em_repr_rl = CRepr;
                       } in
            CMerge(rcl,agemerge,init_po pexp,init_trl l,AgeCache.empty)
          end
        in
        Debug.dprintf2 debug "[Exp] @[need_sem: %a@]@\n" Rlist.Tmp.print l;
        Sem.S.H.add Sem.h v l

  let doexp state (type a) (k : a exp) (exp : a ) age =
    let module Exp = (val (get_exp k) : Exp' with type t = a) in
    Exp.iterexp state age exp


  let need_pexp t (Pexp(age,k,exp,_,_)) =
    (** because explearnt is used for adding the conflict *)
    if not (Exp.equal k explearnt) &&
       not (Exp.equal k expfact) then begin
      check_age t age;
      doexp t k exp age
    end


  let rec find_now t = function
    | Cl(clo0,clr0,(Pexp(age,kexp,exp,_,_concache) as pexp)) -> begin
        match Age.H.find_opt t.lookforcl t.tage with
        | Some trl ->
          Age.H.remove t.lookforcl t.tage;
          let pexpmerge = {
            em_other_cl = clo0;
            em_other_rl = add_needed_cl t (Age.pred t.tage) clo0;
            em_pexp = pexp;
            em_repr_cl = clr0;
            em_repr_rl = add_needed_cl t (Age.pred t.tage) clr0;
          } in
          Debug.dprintf2 debug "[Exp] @[find_now cl %a@]@\n"
            Rlist.Tmp.print_pexpmerge pexpmerge;
          doexp t kexp exp age;
          set_po trl pexpmerge;
        | None -> () end;
      true
    | Sem(cl,sem,v,(Pexp(age,kexp,exp,_,_concache) as pexp)) -> begin
        match Age.H.find_opt t.lookforsem t.tage with
        | Some trl ->
          Age.H.remove t.lookforsem t.tage;
          Debug.dprintf8 debug
            "[Exp] @[find_now sem %a for %a with value %a by %a@]@\n"
            Sem.print sem Cl.print cl (Solver.print_sem sem) v
            print_pexp pexp;
          doexp t kexp exp age;
          let pexpmerge = {
            em_other_cl = cl;
            em_other_rl = CRepr;
            em_pexp = pexp;
            em_repr_cl = cl;
            em_repr_rl = CRepr;
          } in
          set_po trl pexpmerge
        | None -> () end;
      true
    | Dom(cl,dom,(Pexp(age,kexp,exp,_,_concache) as pexp)) ->
      let hcl = Dom.Vector.get t.lookfordom dom in
      let found = Cl.H.find_opt hcl cl in
      begin match found with
      | None -> ()
      | Some () ->
        Cl.H.remove hcl cl;
        Debug.dprintf6 debug "[Exp] @[find_now dom %a for %a by %a@]@\n"
          Dom.print dom Cl.print cl print_pexp pexp;
        doexp t kexp exp age;
        let hcl = Dom.Vector.get t.needed.ndom dom in
        let ages = Cl.H.find_def hcl Age.M.empty cl in
        assert (not (Age.M.mem t.tage ages));
        let ages = Age.M.add t.tage pexp ages in
        Cl.H.replace hcl cl ages
      end;
      true
    | DomL(cl,dom,v,lastlevel,pexp) ->
      Debug.dprintf4 debug "[Exp] @[find_now domL %a for %a@]@\n"
        Dom.print dom Cl.print cl;
      let hcl = Dom.Vector.get t.dom_can_be_at_current dom in
      if Cl.H.mem hcl cl then begin
        assert (Cl.H.find hcl cl);
        ignore(find_now t (Dom(cl,dom,pexp)))
      end;
      begin match v with
        | None -> () (** TODO what about this none?
                         there is an assert in need_dom for it but...*)
        | Some v -> save_dom_at_limit t lastlevel cl dom v
      end;
      Cl.H.replace hcl cl false;
      true
    | Dec _ -> assert (t.tage == (Explanation.last_dec t.etrail));
      false

  let get_needed :
      Solver.t ->
      Explanation.t -> pexp -> rlist needed
    = fun solver t pexp ->
      let tage = current_age t in
      let state = {
        older  = Array.make (Age.to_int tage + 1) [];
        lookforcl  = Age.H.create 10;
        lookforsem = Age.H.create 10;
        tage;
        tclhist = clhist t;
        lookfordom = Dom.Vector.create (Dom.hint_size ());
        needed     = { ncl = Cl.H.create 17;
                       ndom = Dom.Vector.create (Dom.hint_size ());
                       nsem = SemH.create (Sem.hint_size ())
                     };
        solver;
        etrail = Solver.get_trail solver;
        dom_can_be_at_current = Dom.Vector.create (Dom.hint_size ());
      } in

      (** initialization of lookfor *)
      Dom.iter {Dom.iter = (fun dom ->
        Dom.Vector.set state.lookfordom dom (Cl.H.create 10))};
      (** initialization of needed *)
      Dom.iter {Dom.iter = (fun dom ->
        Dom.Vector.set state.needed.ndom dom (Cl.H.create 10))};
      Sem.iter {Sem.iter = (fun (type a) (sem : a sem) ->
      (** initialization of dom_can_be_at_current *)
      Dom.iter {Dom.iter = (fun dom ->
        Dom.Vector.set state.dom_can_be_at_current dom (Cl.H.create 10))};
        let module Sem = (val (Solver.get_sem sem)) in
        let module S = struct
          type k = a
          type v = Rlist.Tmp.tmprlist
          module S = Sem
          let h = S.H.create 5
        end in
        SemH.set state.needed.nsem sem (module S))};

      (** start from the conflict *)
      Debug.dprintf4 debug "[Exp] look at conflict age %a pexp %a@\n"
        Age.print tage print_pexp pexp;
      need_pexp state pexp;

      let add_current_to_look_for t =
        let iter = function
          | FDom (rcl, dom) ->
            Debug.dprintf6 debug "[Exp] @[to look for dom %a of %a at %a@]@\n"
              Dom.print dom Cl.print rcl Age.print t.tage;
            let hcl = Dom.Vector.get t.dom_can_be_at_current dom in
            if Cl.H.find_exn hcl Impossible rcl then
              Cl.H.replace (Dom.Vector.get t.lookfordom dom) rcl ()
              (** otherwise DomL have been previously reached *)
        in
        let current_one = t.older.(Age.to_int t.tage) in
        List.iter iter current_one
      in

      let rec lookdownthetrail state = function
        | [] ->
          assert (Age.equal state.tage (Age.pred Age.min));
        | modif::trail ->
          Debug.dprintf4 debug
            "[Exp] @[lookdownthetrail age %a@ trail %a@]@\n"
            Age.print state.tage print_modif modif;
          Debug.dprintf2 debug
            "[Exp] @[state.dom_can_be_at_current: %a@]@\n"
            print_dom_can_be_at_current state.dom_can_be_at_current;
          add_current_to_look_for state;
          if find_now state modif then begin
            state.tage <- Age.pred state.tage;
            lookdownthetrail state trail
          end
      in

      lookdownthetrail state (trail t);
      (** Check nothing remains *)
      assert (Dom.Vector.fold_initialized
                (fun acc h -> acc && Cl.H.is_empty h)
                true state.lookfordom);
      Rlist.conv state.needed



end
*)
module type Exp = sig

    type t

    val print: t Pp.printer

    val key: t exp

    (* val iterexp: IterExp.t -> age -> t -> unit *)

    val analyse  : ComputeConflict.t -> age ->
      'a con -> t -> 'a rescon

    val expdom : ComputeConflict.t -> age ->
      'b dom -> Cl.t ->
      'a con -> t -> 'a rescon

  end

module RegisterExp (D:Exp) = struct

  let () =
    VExp.inc_size D.key defined_exp;
    assert (if not (VExp.is_uninitialized defined_exp D.key)
      then raise AlreadyRegisteredKey else true);
    let exp = (module D: Exp' with type t = D.t) in
    VExp.set defined_exp D.key exp
end

module type Con = sig

  type t

  val print: t Pp.printer

  val key: t con

  val same_sem:
    ComputeConflict.t -> age -> 'a sem -> 'a ->
    pexp -> Cl.t -> Cl.t -> t rescon

  val propacl : con_iter -> Age.t -> Cl.t -> Cl.t -> t rescon

  val finalize: ComputeConflict.iter -> t Bag.t -> finalized option

end

module RegisterCon (D:Con) = struct

  let () =
    VCon.inc_size D.key defined_con;
    assert (if not (VCon.is_uninitialized defined_con D.key)
      then raise AlreadyRegisteredKey else true);
    let con = (module D: Con' with type t = D.t) in
    VCon.set defined_con D.key con

end


let analyse solver solver_unsat trail pexp =
  (* Debug.dprintf0 debug "[Conflict] 1) find explanation@\n"; *)
  (* let needed = IterExp.get_needed solver t pexp in *)
  Debug.dprintf0 debug "[Conflict] 2) Computation generate conflict@\n";
  let coniter =
    ComputeConflict.compute_con_iter solver solver_unsat trail pexp in
  Debug.dprintf0 debug "[Conflict] 3) Finalize conflict@\n";
  ComputeConflict.finalize coniter

module type Cho = sig
  module Key  : Stdlib.Datatype
  module Data: sig
    type t
    val print: t Pp.printer
  end

  val choose_decision:
    Solver.Delayed.t -> Key.t -> Data.t decdone

  val make_decision:
    Solver.Delayed.t -> dec -> Key.t -> Data.t -> unit

  val analyse:
    ComputeConflict.t -> 'a con -> Key.t -> Data.t -> 'a rescon

  val key: (Key.t,Data.t) cho

end


module VCho = Cho.MkVector
  (struct type ('k,'d,'unedeed) t =
            (module Cho' with type k = 'k and type d = 'd)
   end)

let defined_cho : unit VCho.t = VCho.create 8

let get_cho k =
  assert (if VCho.is_uninitialized defined_cho k
    then raise UnregisteredKey else true);
  VCho.get defined_cho k

let () =
  hash_chogen := (function
      | Explanation.GCho (cho,k) ->
        let module C = (val get_cho cho) in
        Hashcons.combine (Cho.hash cho) (C.Key.hash k)
    );
  equal_chogen :=
    (fun (Explanation.GCho(cho1,k1)) (Explanation.GCho(cho2,k2)) ->
       match Cho.Eq.eq_type cho1 cho2 with
       | Some (Types.Eq, Types.Eq) ->
         let module C = (val get_cho cho1) in
         C.Key.equal k1 k2
       | None -> false
    )


module RegisterCho(C:Cho) = struct

  let () =
    VCho.inc_size C.key defined_cho;
    assert (if not (VCho.is_uninitialized defined_cho C.key)
      then raise AlreadyRegisteredKey else true);
    let module C = struct include C type k = Key.t type d = Data.t end in
    let cho = (module C: Cho' with type k = C.Key.t and type d = C.Data.t) in
    VCho.set defined_cho C.key cho

end

let choose_decision (type k) (type d) d (cho : (k,d) cho) k =
  let module Cho = (val (get_cho cho)) in
  match Cho.choose_decision d k with
  | DecNo -> DecNo
  | DecTodo v ->
    DecTodo (fun d dec -> Cho.make_decision d dec k v)

let make_decision (type k) (type d) d (cho : (k,d) cho) k v dec =
  let module Cho = (val (get_cho cho)) in
  Cho.make_decision d dec k v

let print_pcho fmt = function
  | Pcho(dec,cho,k,d) ->
    let module C = (val get_cho cho) in
    Format.fprintf fmt "Pcho(%a,%a,%a,%a)"
      Age.print (age_of_dec dec)
      Cho.print cho
      C.Key.print k
      C.Data.print d

let print_chogen fmt = function
  | GCho(cho,k) ->
    let module C = (val get_cho cho) in
    Format.fprintf fmt "GCho(%a,%a)"
      Cho.print cho
      C.Key.print k

module ExpCho = struct
  type t = pcho

  let print = print_pcho

  let key = expcho

  let analyse (type a) t _ (con: a con) (Pcho(_,cho,k,v)) =
    let f (type k) (type d) t con (cho : (k,d) cho) k v =
      let module C = (val get_cho cho) in
      C.analyse t con k v in
    f t con cho k v

  let expdom t age _ _ con pcho =
    analyse t age con pcho

end

module EExpCho = RegisterExp(ExpCho)

let fold_rescon_list t f con acc l =
  List.fold_left (fun acc e ->
    let c = ComputeConflict.get_pexp t e.epexp con in
    f acc t c) acc l

let fold_rescon_list_deps t f con acc l =
  List.fold_left (fun (acc,deps) e ->
    let c,deps' = ComputeConflict.get_pexp_deps t e.epexp con in
    let deps = Deps.concat deps' deps in
    f (acc,deps) t c) acc l

let fold_requested f acc t = function
  | GRequested s   -> f acc t s
  | GOther (con,c) -> ComputeConflict.unknown_con t con c; acc

let fold_requested_deps f ((acc,deps) as accdeps) t = function
  | GRequested s   -> f accdeps t s
  | GOther (con,c) -> acc, Deps.add_unknown_con deps con c

module ConFact = struct
  type t = unit
  let print = DUnit.print

  let key = confact

  let finalize _ = assert false (** catched before *)

  let same_sem t _ _ _ pexp2 _ _  =
    let get_con () t = function
      | GRequested ()      -> ()
      | GOther (con,c) -> ComputeConflict.unknown_con t con c
    in
    get_con () t (ComputeConflict.get_pexp t pexp2 confact);
    GRequested ()

  let propacl _  _ _ _ = raise Impossible
  let _propadom _ _ dom cl v =
    Debug.dprintf6 debug
      "[Conflict] @[Error: Ask ConFact for the limit for %a with %a of %a @]@\n"
      Cl.print cl Dom.print dom (Solver.print_dom_opt dom) v;
    raise Impossible

end

module EConFact = RegisterCon(ConFact)

module ExpFact = struct
  type t = unit

  let print = DUnit.print

  let key = expfact

  let analyse (type a) _t _age (con: a con) () =
    mk_confact con


  let expdom t age _ _ con x =
    analyse t age con x


end

module EExpFact = RegisterExp(ExpFact)

module ExpLimit = struct
  type t = explimit

  let print = print_explimit

  let key = explimit

  let analyse (type a) t age (con: a con) limit =
    match limit with
      | LCl (cl,rcl) ->
        let module C = (val (get_con con)) in
        C.propacl t age cl rcl
      | LDom (dom,{modcl;modage;modpexp}) ->
        let f (type b) (k : b exp) (exp: b) =
          let module S = (val (get_exp k) : Exp' with type t = b) in
          S.expdom t modage dom modcl con exp
        in
        let Pexp(_,k,exp,_,_) = modpexp in
        f k exp

  (** explimit only created in conflict and on pexp that are not from
      conflict *)
  let expdom _ _ _ _ _ _ = raise Impossible

end

module EExpLimit = RegisterExp(ExpLimit)

module ExpSameSem = struct
  type t = Solver.exp_same_sem

  let print fmt = function
    | Solver.ExpSameSem(pexp,cl1,clsem2) ->
      Format.fprintf fmt "ExpSameSem(%a,%a,%a:%a)"
        print_pexp pexp
        Cl.print cl1
        Cl.print (ClSem.cl clsem2)
        ClSem.print clsem2


  let key = Solver.exp_same_sem
(*
  let iterexp t age = function
    | Solver.ExpSameSem(pexp,_,_,k,sem) ->
      IterExp.need_sem t age k sem;
      IterExp.need_pexp t pexp
*)
  let analyse (type b) t age (con : b con) = function
    | Solver.ExpSameSem(pexp,cl1,clsem2) ->
      assert (Ty.equal (Cl.ty cl1) (ClSem.ty clsem2));
      let module C = (val get_con con) in
      match Only_for_solver.sem_of_cl clsem2 with
      | Only_for_solver.Sem(sem,v) ->
        C.same_sem t age sem v pexp cl1 (ClSem.cl clsem2)

  (** not used for dom *)
  let expdom _ _ _ _ _ _ = raise Impossible

end

module EExpSameSem = RegisterExp(ExpSameSem)

module ExpLearnt = struct
  type t = explearnt

  let print fmt = function
    | ExpLearnt(tags) ->
      Format.fprintf fmt "ExpLearnt(%a)"
        Tags.print tags

  let key = explearnt

  let analyse (type a) t _age (con: a con) (ExpLearnt(tags)) =
    t.cdeps <- Deps.add_tags t.cdeps tags;
    return con ConFact.key ()

  let expdom _ _ _ _ con _ =
    return con ConFact.key ()

end

module EExpLearnt = RegisterExp(ExpLearnt)

let check_initialization () =
  let well_initialized = ref true in

  Exp.iter {Exp.iter = fun exp ->
    if VExp.is_uninitialized defined_exp exp then begin
      Format.eprintf
        "[Warning] The explanation %a is not registered@." Exp.print exp;
      well_initialized := false;
    end};

  Con.iter {Con.iter = fun con ->
    if VCon.is_uninitialized defined_con con then begin
      Format.eprintf
        "[Warning] The conflicts %a is not registered@." Con.print con;
      well_initialized := false;
    end};

  !well_initialized

let print_conflicts = Debug.register_info_flag
  ~desc:"for@ the@ printing@ of@ the@ conflicts@ found."
  "conflicts"


let print_decision = Debug.register_info_flag
  ~desc:"for@ the@ printing@ of@ the@ decisions@ done."
  "decisions"

(** Helpers *)
module ChoGenH = ChoH


(** Exception printer *)
let () = Exn_printer.register(fun fmt exn ->
    match exn with
    | DontKnowThisDec(dom,cl) ->
      Format.fprintf fmt "DontKnowThisDec(%a,%a)"
        Dom.K.print dom Cl.print cl
    | exn -> raise exn
  )
