(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Popop
open Types
open Explanation

type 'a rescon = 'a Explanation.rescon =
| GRequested: 'a -> 'a rescon
| GOther: 'b con * 'b ->  'a rescon

val return: 'a con -> 'b con -> 'b -> 'a rescon
val return_nothing: 'a con -> 'a rescon
(*
module AgeCache: sig
  type t
  type 'a cache
  val create_cache: string -> 'a cache
  val no_cache: unit cache
end
*)
(** An information can depend on a list of merge *)

type edge = {
  eto : Cl.t; (** equals to *)
  epexp : pexp (** explication for this merge *);
  einv  : bool; (** which way was pexp *)
}

val print_edge: edge Pp.printer

type rlist = edge list

val print_rlist: rlist Pp.printer
(*
module IterExp : sig
  exception AgeTooYoung

  type t

  val need_equal  : t -> age -> Cl.t -> Cl.t -> unit
  val need_dom    : t -> age -> Cl.t -> 'a dom -> unit
  val need_sem    : t -> 'a sem -> 'a -> unit
  val need_pexp   : t -> pexp -> unit

end
*)
exception NotMarkedExp

module ComputeConflict: sig
  type t

  val before_first_dec: t -> Age.t -> bool

  val get_equal: t -> age -> Cl.t -> Cl.t -> rlist
  (** get the list of merge that make the too classes equal. *)
  val get_repr_at: t -> age -> Cl.t -> Cl.t
  (** get the repr of the class at this age . *)

  val get_dom    : t -> age -> Cl.t -> 'b dom -> mod_dom list
  (** get the explications for the domain of this class at the given age *)

  val get_dom_path    : t -> age -> Cl.t -> 'b dom -> mod_dom list
  (** get the explications for the domain of this class at the given age that
      only applied on this cl *)

  val get_dom_last    : t -> age -> Cl.t -> 'b dom -> mod_dom
  (** get the last explication for the domain of this class at the
      given age that only applied on this cl *)

  val get_pexp_deps : t -> pexp -> 'a con -> 'a rescon * Deps.t
  val add_deps: t -> Deps.t -> unit
  val get_current_deps: t -> Deps.t
  (** mostly for debugging or assert checking *)

  val get_pexp     : t -> pexp -> 'a con -> 'a rescon
  (** same as get_pexp_deps followed by add_deps *)

  val unknown_con: t -> 'b con -> 'b -> unit
  (** same as [Deps.add_unknown_con Deps.empty con a] followed by add_deps *)


  val set_dec_cho : t -> ('k,'d) cho -> 'k -> unit
  val set_dec_cho_cond : t -> 'a dom -> Cl.t -> ('k,'d) cho -> 'k -> unit

  type iter
  val dep_dec: iter -> 'a dom -> Cl.t -> unit

end

val fold_requested:
  ('b -> ComputeConflict.t -> 'a -> 'b) ->
  'b -> ComputeConflict.t -> 'a rescon -> 'b
(** The other case is applied to {!ComputeConflict.unknown_con} *)

val fold_requested_deps:
  ('b * Deps.t -> ComputeConflict.t -> 'a -> 'b * Deps.t) -> 'b * Deps.t ->
  ComputeConflict.t -> 'a rescon -> 'b * Deps.t
(** The other case is applied to {!ComputeConflict.unknown_con} *)

val fold_rescon_list:
  ComputeConflict.t -> ('a -> ComputeConflict.t -> 'c rescon -> 'a)
  -> 'c con -> 'a -> rlist -> 'a

val fold_rescon_list_deps:
  ComputeConflict.t ->
  ('a * Deps.t -> ComputeConflict.t -> 'c rescon -> 'a * Deps.t)
  -> 'c con -> 'a * Deps.t -> rlist -> 'a * Deps.t

(* val fold_rescon_list_deps_cache: *)
(*   cache:(('a * Deps.t) AgeCache.cache) -> *)
(*   concat:('a -> 'a -> 'a) -> *)
(*   empty:'a -> *)
(*   ComputeConflict.t -> *)
(*   (ComputeConflict.t -> 'c rescon -> 'a * Deps.t) *)
(*   -> Age.t -> 'c con -> rlist -> 'a * Deps.t *)


exception DontKnowThisCon


module type Exp = sig

  type t

  val print: t Pp.printer

  val key: t exp

  (* val iterexp: IterExp.t -> age -> t -> unit *)

  val analyse  :
    ComputeConflict.t -> age -> 'a con -> t -> 'a rescon

  val expdom : ComputeConflict.t -> age ->
    'b dom -> Cl.t ->
    'a con -> t -> 'a rescon

end

module RegisterExp(E:Exp) : sig end

type testconflict =
  | False (* false in this context *)
  | True (* true in this context *)
  | ToDecide (* can be decided *)

type 'a fold_decisions =
  {fold_decisions : 'k 'd. ('a -> ('k,'d) cho -> 'k -> 'd -> 'a) }

class type finalized = object
  method print: Format.formatter -> unit
  method conflict_add: Solver.Delayed.t -> bool Cl.M.t
  method test: Solver.Delayed.t -> testconflict
  method decide: 'a. 'a fold_decisions -> Solver.Delayed.t -> 'a -> 'a
    (** to call only when test is [Todecide] *)
end

val print_finalized: finalized Pp.printer

module type Con = sig

  type t

  val print: t Pp.printer

  val key: t con

  (** Particular conflict computation *)
  val same_sem:
    ComputeConflict.t -> age -> 'a sem -> 'a ->
    pexp -> Cl.t -> Cl.t -> t rescon

  val propacl : ComputeConflict.t -> Age.t -> Cl.t -> Cl.t -> t rescon
    (* cl1 -> cl2 repr *)

  (** Finalization and propagation of the conflict *)
  val finalize: ComputeConflict.iter -> t Bag.t -> finalized option
      (** None if the conflict is empty (always satisfied constraint)
          TODO better doc we produce the conflcit ro the explication?
      *)



end

module RegisterCon(C:Con) : sig end

type 'd decdone  =
| DecNo (** no decision to do *)
| DecTodo of 'd (** this decision should be done *)

module type Cho = sig
  module Key  : Stdlib.Datatype
  module Data: sig
    type t
    val print: t Pp.printer
  end

  val choose_decision:
    Solver.Delayed.t -> Key.t -> Data.t decdone
  (** Answer the question: Is the decision still needed? *)

  val make_decision:
    Solver.Delayed.t -> dec -> Key.t -> Data.t -> unit
  (** If it is ({!DecTodo} returned by choose_decision) the decision
      can be done using the usual {!Solver.Delayed} functions *)

  val analyse:
    ComputeConflict.t -> 'a con -> Key.t -> Data.t -> 'a rescon
  (** If a conflict happened when this decision is the last. This
      function is called *)

  val key: (Key.t,Data.t) cho

end

module RegisterCho(C:Cho) : sig end


val analyse:
  Solver.t -> (** State of the solver before the last dec *)
  Solver.t -> (** State of the solver unsat *)
  Explanation.t -> pexp -> (* 'a must be a conflict *)
  finalized option * tags * chogen Bag.t
    (* maximum age not in conflict *)

val choose_decision:
  Solver.Delayed.t -> ('k,'d) cho -> 'k ->
  (Solver.Delayed.t -> dec -> unit) decdone


val make_decision:
  Solver.Delayed.t -> ('k,'d) cho -> 'k -> 'd -> dec -> unit

val print_pexp: pexp Pp.printer
val print_chogen: Explanation.chogen Pp.printer

val check_initialization: unit -> bool
(** Check if the initialization of all the exp, con have been done *)

val confact: unit con

type explimit =
  | LCl: Cl.t * Cl.t -> explimit
  (** At the limit the first cl is represented by this cl *)
  | LDom: 'a dom * Explanation.mod_dom -> explimit
  (** At the limit this class have this value for this dom *)

val is_explimit : pexp -> explimit option

(** Stuff common to Boolean *)

val dom_bool: bool dom
val conclause: bool Cl.M.t Explanation.con

val print_conflicts: Debug.flag
val print_decision: Debug.flag

(** Helpers for scheduler *)
module ChoGenH : Exthtbl.Hashtbl.S with type key = chogen

type explearnt =
| ExpLearnt: Tags.t -> explearnt

val explearnt : explearnt exp
