(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)
open Popop
open Types

val equality    : Cl.t list -> Cl.t
val disequality : Cl.t list -> Cl.t

val is_equal    : Solver.d -> Cl.t -> Cl.t -> bool
val is_disequal : Solver.d -> Cl.t -> Cl.t -> bool

val ite : Cl.t -> Cl.t -> Cl.t -> Cl.t

val th_register : Solver.d -> unit

open Explanation

type special_equality = {
  equality:
    Conflict.ComputeConflict.t ->
    Explanation.Age.t -> Cl.t -> Cl.t -> unit;
  disequality:
    Conflict.ComputeConflict.t -> Explanation.Age.t ->
    hyp:bool -> Cl.t -> Cl.t -> Cl.t -> Cl.t -> unit;
  merged:
    Conflict.ComputeConflict.t ->
    Explanation.Deps.t -> Explanation.Age.t ->
    Cl.t -> Cl.t ->
    Explanation.pexp ->
    Cl.t -> Cl.t -> Explanation.Deps.t;
  repr:
    Conflict.ComputeConflict.t -> Explanation.Deps.t -> Explanation.Age.t ->
    Cl.t -> Conflict.rlist ->
    Explanation.Deps.t;
  dodec: bool;
}

val register_sort: Ty.t -> special_equality -> unit
val register_sort_con: Ty.t -> dodec:bool -> 'a con -> unit

module GenEquality: sig
  val equality: Conflict.ComputeConflict.t ->
    Explanation.Age.t -> Types.Cl.t -> Types.Cl.t -> Ty.t -> unit

end

type eqconflict =
  | Eq : Cl.t * Cl.t * bool -> eqconflict

val choequal : (eqconflict, unit) Explanation.cho
