(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

type t

val run:
  ?nodec:unit ->
  theories:(Solver.d -> unit) list ->
  (Solver.d -> unit) ->
  [`Contradiction | `Done of Solver.d]

exception Contradiction

val run_exn:
  ?nodec:unit ->
  theories:(Solver.d -> unit) list ->
  (Solver.d -> unit) ->
  Solver.d


(** {2 Expert interface} *)
val new_solver: unit -> t
(** Create a scheduler *)

val get_t: t -> Solver.t
(** Get the solver. *)

val get_delayed   : t -> Solver.Delayed.t
(** Get the delayed solver. *)

val flush_delayed : t -> unit
(** Flush the delayed solver. Apply all the delayed actions.
    Previous {!get_delayed} results shouldn't be used anymore.
*)

val stop_delayed  : t -> unit

exception NeedStopDelayed

val run_inf_step: ?nodec:bool -> t -> unit
(** Raise NeedStopDelayed if {!stop_delayed} have not been called since the last
    {!get_delayed}.
*)

val new_handler : t -> t
(** Raise NeedStopDelayed if {!stop_delayed} have not been called since the last
    {!get_delayed}.
*)

