open OUnit
open Tests_lib


let unused = Template.lazy_propagation

let opt_seed = ref 0

let print_seed fmt = function
  | None -> Format.fprintf fmt "No"
  | Some [|i|] -> Format.fprintf fmt "%i" i
  | _ -> assert false

let make_tests acc seed =
  let module Utils = Tests_utils in
  let module Uf = Tests_uf in
  let module Arith = Tests_arith in
  let module UfArith = Tests_arith_uninterp in
  let module Bv = Tests_bv in
  let module Bool = Tests_bool in
  let test = ((Pp.sprintf "seed %a" print_seed seed) >:::
                 [Utils.tests;
                  Uf.tests ; Arith.tests; UfArith.tests; Bool.tests; Bv.tests]) in
  let test = test_decorate
    (fun f -> (fun () -> Shuffle.set_shuffle seed; f ())) test in
  test::acc

let tests () =
  let l = Util.foldi (fun acc i -> make_tests acc (Some [|i|])) []
    (!opt_seed + 1) (!opt_seed + 9)in
  make_tests l None

let tests () =
(*  test_decorate
    (fun f ->
        if Printexc.backtrace_status ()
        then fun () -> try f () with exn -> Printexc.print_backtrace stdout;
          raise exn
        else f
    ) *)
    TestList (tests ())

(** From oUnit.ml v 1.2.2 *)
(** just need to make the tests lazily computed *)

(* Returns true if the result list contains successes only *)
let rec was_successful =
  function
    | [] -> true
    | RSuccess _::t
    | RSkip _::t ->
        was_successful t

    | RFailure _::_
    | RError _::_
    | RTodo _::_ ->
        false


(* Call this one from you test suites *)
let run_test_tt_main ?(arg_specs=[]) suite =
  let only_test = ref [] in
  let () =
    Arg.parse
      (Arg.align
         [
           "-only-test",
           Arg.String (fun str -> only_test := str :: !only_test),
           "path Run only the selected test";

           "-list-test",
           Arg.Unit
             (fun () ->
                List.iter
                  (fun pth ->
                     print_endline (string_of_path pth))
                  (test_case_paths (suite ()));
                exit 0),
           " List tests";
         ] @ arg_specs
      )
      (fun x -> raise (Arg.Bad ("Bad argument : " ^ x)))
      ("usage: " ^ Sys.argv.(0) ^ " [-verbose] [-only-test path]*")
  in
  let () = Debug.Args.set_flags_selected () in
  let verbose = Debug.test_flag debug in
  let nsuite =
    if !only_test = [] then
      suite ()
    else
      begin
        match test_filter ~skip:true !only_test (suite ()) with
          | Some test ->
              test
          | None ->
              failwith ("Filtering test "^
                        (String.concat ", " !only_test)^
                        " lead to no test")
      end
  in
  let result = run_test_tt ~verbose nsuite in
    if not (was_successful result) then
      exit 1
    else
      result

(*** End *)

let () =
  if not (Solver.check_initialization ()) ||
     not (Conflict.check_initialization ()) then
    exit 1

let _ =
  try
    run_test_tt_main
      ~arg_specs:["--seed",Arg.Set_int opt_seed,
                  " Base seed used for shuffling the arbitrary decision";
                  Debug.Args.desc_debug;
                  Debug.Args.desc_debug_all]
      tests
  with e when not (Debug.test_flag Debug.stack_trace) ->
    Format.eprintf "%a@." Exn_printer.exn_printer e;
    exit 1
