open OUnit
open Popop

let debug = Debug.register_flag
  ~desc:" Run the test in verbose mode." "ounit"

let test_file s =
  s >::
    fun () ->
      let d = Popop_of_altergo.read_file s in
      assert_bool "" (Popop_of_altergo.Valid = Popop_of_altergo.check_goal d)


let test_split s =
  s >::: begin
    let l = Popop_of_altergo.read_split s in
    List.map (fun (res_attended,loc,d) ->
        TestCase (fun () ->
          let s =  Pp.string_of_wnl Loc.gen_report_position loc in
          let actual_res =
            Loc.with_this_loc Popop_of_altergo.check_goal loc d in
          let b = (Popop_of_altergo.Valid = actual_res) in
          match res_attended with
          | "Valid" -> assert_bool s b
          | "Unknown" -> assert_bool s (not b)
          | _ -> invalid_arg "Unknown status")

    ) l
  end

let tests_smt2 expected dir =
  let files = Sys.readdir dir in
  Array.sort String.compare files;
  let files = Array.to_list files in
  List.map
    (fun s ->
      s >: TestCase (fun () ->
        let cmds = Popop_of_smtlib2.read_file (Filename.concat dir s) in
        let res = Popop_of_smtlib2.check_file cmds in
        Debug.dprintf3 debug "@[%s: %a@]@\n"
          s Popop_of_smtlib2.print_status res;
        assert_bool s (res = expected);
      )) files

let (&:) s l = s >::: (List.map (fun f -> TestCase f) l)


let ty_ctr = Types.Ty.Constr.create "a"
let ty = Types.Ty.ctr ty_ctr


let register d cl =
  Solver.Delayed.register d cl;
  Solver.Delayed.flush d

let merge d cl1 cl2 =
  Solver.Delayed.merge d Explanation.pexpfact cl1 cl2;
  Solver.Delayed.flush d

let is_equal = Solver.Delayed.is_equal
