

let rec new_problem f =
  let filename,cout = Filename.open_temp_file "search_diff" ".cnf" in
  let fd = Unix.descr_of_out_channel cout in
  let pid = Unix.create_process "fuzzsat" Sys.argv Unix.stdin fd fd in
  let pid', status = Unix.wait () in
  assert (pid = pid');
  match status with
  | Unix.WEXITED 0 ->
    f filename;
    close_out cout;
    Sys.remove filename
  | _ ->
    Format.printf "Error in fuzzsat:@.%t@."
      (Sysutil.file_contents_fmt filename);
    close_out cout;
    Sys.remove filename;
    new_problem f


let sat_for_minisat filename =
  let cmd =
    Printf.sprintf "minisat %s | grep -q \"^SATISFIABLE\"" filename in
  Sys.command cmd = 0

type result = Error | Result of bool

let sat_for_popop filename =
  let error,cout = Filename.open_temp_file "popop" ".out" in
  let fd = Unix.descr_of_out_channel cout in
  let pid = Unix.create_process "./popop.native" [|"./popop.native";filename|]
      Unix.stdin fd fd in
  let pid', _status = Unix.wait () in
  assert (pid = pid');
  let cmd = Printf.sprintf "grep -q \"anomaly\" %s" error in
  if Sys.command cmd = 0 then
    begin
    Format.printf "Error in popop:@.%t@."
      (Sysutil.file_contents_fmt error);
    close_out cout;
    Sys.remove error;
    Error
    end
  else let cmd = Printf.sprintf "grep -q \"^Sat\" %s" error in
    let r = (Sys.command cmd = 0) in
    close_out cout;
    Sys.remove error;
    Result r

let feedback reason filename =
  let basename = Filename.basename filename in
  Sysutil.backup_file basename;
  Sysutil.copy_file filename basename;
  Printf.printf "%s on %s\n%!" reason basename

let rec run () =
  new_problem (fun filename ->
      match sat_for_minisat filename, sat_for_popop filename with
      | _, Error -> feedback "anomaly" filename
      | b1, Result b2 when b1 <> b2 -> feedback "different" filename;
        Printf.printf "minisat sat:%b popop sat:%b\n%!" b1 b2;
        ignore (Sys.command (Printf.sprintf "wc -l %s" filename))
      | _ -> Printf.printf ".%!");
  run ()

let () = run ()
